@@ -16,6 +16,7 @@
 
 from kazoo import client
 from kazoo import exceptions
+from kazoo import security
 try:
     from kazoo.handlers import eventlet as eventlet_handler
 except ImportError:
@@ -56,7 +57,9 @@ class ZooKeeperLock(locking.Lock):
                                    "operation error: %s" % (e),
                                    cause=e)
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
+        if shared:
+            raise tooz.NotImplemented
         blocking, timeout = utils.convert_blocking(blocking)
         return self._lock.acquire(blocking=blocking,
                                   timeout=timeout)
@@ -133,7 +136,7 @@ class KazooDriver(coordination.CoordinationDriverCachedRunWatchers):
     """
 
     def __init__(self, member_id, parsed_url, options):
-        super(KazooDriver, self).__init__(member_id)
+        super(KazooDriver, self).__init__(member_id, parsed_url, options)
         options = utils.collapse(options, exclude=['hosts'])
         self.timeout = int(options.get('timeout', '10'))
         self._namespace = options.get('namespace', self.TOOZ_NAMESPACE)
@@ -447,6 +450,18 @@ class KazooDriver(coordination.CoordinationDriverCachedRunWatchers):
         # Creates a kazoo client,
         # See: https://github.com/python-zk/kazoo/blob/2.2.1/kazoo/client.py
         # for what options a client takes...
+        if parsed_url.username and parsed_url.password:
+            username = parsed_url.username
+            password = parsed_url.password
+
+            digest_auth = "%s:%s" % (username, password)
+            digest_acl = security.make_digest_acl(username, password, all=True)
+            default_acl = (digest_acl,)
+            auth_data = [('digest', digest_auth)]
+        else:
+            default_acl = None
+            auth_data = None
+
         maybe_hosts = [parsed_url.netloc] + list(options.get('hosts', []))
         hosts = list(compat_filter(None, maybe_hosts))
         if not hosts:
@@ -458,6 +473,8 @@ class KazooDriver(coordination.CoordinationDriverCachedRunWatchers):
             'connection_retry': options.get('connection_retry'),
             'command_retry': options.get('command_retry'),
             'randomize_hosts': strutils.bool_from_string(randomize_hosts),
+            'auth_data': auth_data,
+            'default_acl': default_acl,
         }
         handler_kind = options.get('handler')
         if handler_kind:
